package com.jz.dm.filter;

import com.alibaba.fastjson.JSONObject;
import com.baomidou.mybatisplus.core.conditions.update.UpdateWrapper;
import com.jz.common.utils.HttpsUtils;
import com.jz.common.utils.RedisUtils;
import com.jz.dm.common.constant.Constants;
import com.jz.dm.common.constant.TagConstants;
import com.jz.dm.common.enums.GatewayResultCode;
import com.jz.dm.common.enums.apiInterface.ApiStatusEnum;
import com.jz.dm.common.enums.auth.AuthReqTypeEnum;
import com.jz.dm.common.exception.GatewayException;
import com.jz.dm.common.util.DateUtil;
import com.jz.dm.common.util.LogUtil;
import com.jz.dm.mapper.ApiInterfaceMapper;
import com.jz.dm.models.domian.ApiAuth;
import com.jz.dm.models.domian.ApiInterface;
import com.jz.dm.models.domian.ApiOrg;
import com.jz.dm.models.enity.GatewayRequest;
import com.jz.dm.models.enity.GatewayResponse;
import com.jz.dm.service.ApiInterfaceService;
import com.jz.dm.service.AuthService;
import com.jz.dm.service.OrganizationManageService;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

import javax.annotation.Resource;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;

/**
 * @author ZC
 * @PACKAGE_NAME: com.jz.dm.filter
 * @PROJECT_NAME: jz-dm-parent
 * @NAME: AuthFilter
 * @DATE: 2020-12-23/17:43
 * @DAY_NAME_SHORT: 周三
 * @Description: 认证过滤器
 **/
@Component("authFilter")
@Slf4j
public class AuthFilter extends AbstractFilter {
    private static final Logger LOGGER = LoggerFactory.getLogger(AuthFilter.class);

    @Autowired
    private AuthService authService;
    @Autowired
    private OrganizationManageService organizationManageService;
    @Autowired
    private ApiInterfaceService apiInterfaceService;
    @Resource
    private ApiInterfaceMapper apiInterfaceMapper;
    @Autowired
    private RedisUtils redisUtils;
    @Autowired
    private HttpsUtils httpUtils;
    @Value("${data.bank.balanceUrl}")
    private String balanceUrl;

    @Override
    public String getFilterName() {
        return "AuthFilter";
    }

    @Override
    public int getOrder() {
        return Constants.FILTER_ORDER_5;
    }

    @Override
    protected void internalDoFilter(GatewayRequest request, GatewayResponse response,
                                    FilterChain chain) {
        try {
            ApiInterface apiInterface = apiInterfaceService.getApiInfo(request.getApiKey());
            if (null == apiInterface) {
                throw new GatewayException(GatewayResultCode.ILLEGAL_REQUEST);
            }
            //下架状态$$ 检查是否有有效调用的api,如果有就放行,没有就置为无效
            if (ApiStatusEnum.SOLDOUT.name().equals(apiInterface.getStatus())) {
                handleApiStatus(apiInterface);
            } else {//非下架状态
                if (!ApiStatusEnum.ISSUE.name().equals(apiInterface.getStatus())) {
                    throw new GatewayException(GatewayResultCode.API_STATUS_EXCEPTION);
                }
            }
            JSONObject parameter = JSONObject.parseObject(request.getParams());
            if (null == parameter) {
                throw new GatewayException(GatewayResultCode.REQUEST_PARAM_EMPTY);
            }
            Boolean isTest = parameter.getBoolean("isTest");
            if (!isTest) {//是否测试数据
                String authCode = parameter.getString("authCode");
                ApiAuth authAuth = authService.getAuthUser(authCode, apiInterface.getId());
                if (null == authAuth) {
                    throw new GatewayException(GatewayResultCode.REQUEST_NOT_AUTH);
                }
                if ("1".equals(authAuth.getHandler())) {//处理状态,已经调用成功后授权码失效
                    throw new GatewayException(GatewayResultCode.ILLEGAL_REQUEST);
                }
                //查询认证组织信息
                ApiOrg apiOrg = organizationManageService.getAuthOrganization(authAuth.getApiOrgId());
                if (null == apiOrg) {
                    throw new GatewayException(GatewayResultCode.ORG_NOT_EXIST);
                }
                //状态检查(只限制按次请求$$ 不限制按时间请求)
                if (!apiOrg.getStatus()) {
                    throw new GatewayException(GatewayResultCode.ORG_STATE_EXCEPTION);
                }
                checkBill(request, authCode, authAuth);
            }
            chain.doFilter(request, response);
        } catch (Exception e) {
            if (e instanceof GatewayException) {
                throw (GatewayException) e;
            }
            log.error("~~~~~~~~~~~~~~~~认证过滤异常~~~~~~~~~~~~~~~~~");
            log.error("异常信息:{}", e.getMessage());
            LogUtil.error(LOGGER, e,
                    "authFilter doFilter error. response=" + response.getResponse());
            response.clearAttributes();
            response.setCode(GatewayResultCode.UNKNOWN_EXCEPTION.getCode());
            response.setMsg(GatewayResultCode.UNKNOWN_EXCEPTION.getMsg());
        }

    }

    /**
     * 校验计费信息
     *
     * @param request
     * @param authCode
     * @param authAuth
     * @throws ParseException
     */
    private void checkBill(GatewayRequest request, String authCode,
                           ApiAuth authAuth) throws ParseException {
        switch (authAuth.getAuthMode()) {
            case "POWER_CALL_MODE": //按次调用
                //查询数据银行银行余额是否充足
                getDataAmountResult(request);
                try { //记录请求次数(每天限制请求次数)
                    String limitKey = TagConstants.OPEN_API_REDIS_LIMIT_KEY + authCode;
                    Integer reqValue = (Integer) redisUtils.getObj(limitKey);
                    long timeOut = 0;
                    if (AuthReqTypeEnum.DAY.name().equals(authAuth.getReqType())) { //按天
                        timeOut = DateUtil.calculateNowResidueTime();
                    } else if (AuthReqTypeEnum.MONTH.name().equals(authAuth.getReqType())) {//按月
                        timeOut = DateUtil.calculateNowMonthTime();
                    } else if (AuthReqTypeEnum.YEAR.name().equals(authAuth.getReqType())) {//按年
                        //暂时不支持年
                    }
                    if (null != reqValue) {
                        /*   Integer value = Integer.valueOf(reqValue);*/
                        if (reqValue > authAuth.getReqFrequency()) {//超出最大请求次数
                            throw new GatewayException(GatewayResultCode.REQUEST_LIMIT_EXCEPTION);
                        } else if (reqValue <= authAuth.getReqFrequency()) {
                            redisUtils.delAndAdd(limitKey, limitKey, reqValue + 1, timeOut);
                        }
                    } else {
                        redisUtils.set(limitKey, 1, timeOut);
                    }
                } catch (Exception ex) {
                    log.error("~~~~~~~~~~~~~~~~~~~~~~过滤计次信息异常~~~~~~~~~~~~~~~~~~~");
                    log.error("异常信息:{}", ex.getMessage());
                    throw new GatewayException(GatewayResultCode.REQUEST_LIMIT_EXCEPTION);
                }
                break;
            case "RECORD_TIME_MODE": //按时间调用
                SimpleDateFormat dateFormat = new SimpleDateFormat(Constants.DATE_TIME_FORMAT);
                String formatDate = dateFormat.format(new Date());
                Date currentDate = dateFormat.parse(formatDate);
                Date validEndTime = authAuth.getValidEndTime();
                //如果 date1 在 date2 之前，before 返回 true，否则返回 false
                if (currentDate.after(validEndTime)) {//超出时间
                    updateCallTimeStatus(authAuth);//查看是否为已处理状态
                    throw new GatewayException(GatewayResultCode.RATE_LIMIT_EXCEEDED);
                }
                break;
            case "PERMANENT_TIME_MODE"://永久有效(直接跳出)
                break;
            default:
                throw new GatewayException(GatewayResultCode.ILLEGAL_REQUEST);
        }
    }

    /**
     * 更新调用时间状态
     */
    private void updateCallTimeStatus(ApiAuth authAuth) {
        if (null != authAuth && "0".equals(authAuth.getHandler())) {
            if (authService.updateApiAuthStatus(authAuth) == 0) {
                log.info("------------------更新状态调用状态异常---------------");
                throw new GatewayException(GatewayResultCode.UNKNOWN_EXCEPTION);
            }
        }
    }

    /**
     * 按次调用检查余额是否充足
     *
     * @param request
     */
    private void getDataAmountResult(GatewayRequest request) {
        JSONObject params = JSONObject.parseObject(request.getParams());
        Integer assetsId = params.getInteger("assetsId");
        Integer userId = params.getInteger("userId");
        String dataPrice = params.getString("dataPrice");
        if (null == assetsId || null == userId || StringUtils.isBlank(dataPrice)) {
            throw new GatewayException(GatewayResultCode.ILLEGAL_ARGUMENT);
        }
        if (null != params) {
            JSONObject pObject = new JSONObject();
            pObject.put("assetsId", assetsId);
            pObject.put("userId", userId);
            pObject.put("dataPrice", dataPrice);
            String respResult = httpUtils.submitPost(balanceUrl + "/mall/financeCustomerAssets/findAssets", pObject.toString());
            JSONObject result = JSONObject.parseObject(respResult);
            if (null != result) {
                if (200 != result.getInteger("code")) {
                    log.info("~~~~~~~~~~~~~~~~~~异常信息为:{}", result.getString("message"));
                    throw new GatewayException(GatewayResultCode.CALL_AMOUNT_NOT_ENOUGH);
                }
            }
        }
    }

    /**
     * 处理Api状态信息
     *
     * @param apiInterface
     */
    private void handleApiStatus(ApiInterface apiInterface) {
        //检查是否有有效的认证的API
        Integer authStatus = apiInterfaceService.getValidAuthStatus(apiInterface);
        if (null == authStatus || authStatus == 0) {
            UpdateWrapper<ApiInterface> updateWra = new UpdateWrapper<>();
            updateWra.eq("status", ApiStatusEnum.EXPIRY);
            updateWra.eq("is_deleted", 1);
            if (apiInterfaceMapper.update(null, updateWra) == 0) {
                log.info("------------------更新API状态失败-----------");
                throw new GatewayException(GatewayResultCode.FAIL);
            }
        }
    }
}
