package com.jz.dm.filter;

import com.alibaba.fastjson.JSONObject;
import com.jz.common.utils.Md5;
import com.jz.dm.common.constant.Constants;
import com.jz.dm.common.enums.GatewayResultCode;
import com.jz.dm.common.exception.GatewayException;
import com.jz.dm.common.exception.SignatureException;
import com.jz.dm.common.util.LogUtil;
import com.jz.dm.common.util.MapUtil;
import com.jz.dm.models.domian.ApiAuth;
import com.jz.dm.models.enity.GatewayRequest;
import com.jz.dm.models.enity.GatewayResponse;
import com.jz.dm.service.ApiInterfaceService;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

/**
 * @Description: 验签过滤器(验证签名信息)
 * @Author: Mr.zhang
 * @Date: 2021-1-6
 */
@Component
public class VerifySignFilter extends AbstractFilter {
    private static final Logger LOGGER = LoggerFactory.getLogger(VerifySignFilter.class);
    @Autowired
    private ApiInterfaceService apiInterfaceService;

    @Override
    public int getOrder() {
        return Constants.FILTER_ORDER_3;
    }

    @Override
    public String getFilterName() {
        return "VerifySignFilter";
    }

    @Override
    protected void internalDoFilter(GatewayRequest request, GatewayResponse response,
                                    FilterChain chain) {
        try {
            JSONObject jsonObject = JSONObject.parseObject(request.getParams());
            if (!jsonObject.getBoolean("isTest")) {//是否测试调用
                //对签约参数进行字典排序
                String signParams = MapUtil.getSignValue(request.getApiKey(), request.getMethod(), request.getSignType());
                if (StringUtils.isNotBlank(signParams)) {
                    String authCode = jsonObject.getString("authCode");
                    //需要传入授权码
                    ApiAuth apiAuthInfo = apiInterfaceService.getApiAuthInfo(request.getApiKey(), authCode);
                    if (null == apiAuthInfo) {
                        throw new GatewayException(GatewayResultCode.ILLEGAL_REQUEST);
                    }
                    String sign = Md5.encrypt(signParams, apiAuthInfo.getSalt());
                    if (!request.getSign().equals(sign)) {
                        throw new GatewayException(GatewayResultCode.SIGN_ERROR);
                    }
                } else {
                    throw new GatewayException(GatewayResultCode.SIGN_ERROR);
                }
            }
            chain.doFilter(request, response);
        } catch (SignatureException ex) {
            LogUtil.error(LOGGER, ex,
                    "sign response error. response=" + response.getResponse());
            response.clearAttributes();
            response.setCode(ex.getResultCode().getCode());
            response.setMsg(ex.getResultCode().getMsg());
        } catch (Throwable ex) {
            if (ex instanceof GatewayException) {
                throw (GatewayException) ex;
            }
            LogUtil.error(LOGGER, ex,
                    "signatureFilter doFilter error. response=" + response.getResponse());
            response.clearAttributes();
            response.setCode(GatewayResultCode.UNKNOWN_EXCEPTION.getCode());
            response.setMsg(GatewayResultCode.UNKNOWN_EXCEPTION.getMsg());
        }
    }

}
