package com.jz.dm.service.request;

import com.alibaba.csp.sentinel.annotation.SentinelResource;
import com.alibaba.fastjson.JSONObject;
import com.jz.common.utils.HttpsUtils;
import com.jz.common.utils.RedisUtils;
import com.jz.dm.common.constant.LoggingConstants;
import com.jz.dm.common.enums.GatewayResultCode;
import com.jz.dm.common.enums.apiInterface.ApiInfoOutTypeEnum;
import com.jz.dm.common.enums.auth.AuthModeEnum;
import com.jz.dm.common.exception.GatewayException;
import com.jz.dm.common.util.OpenApiRequest;
import com.jz.dm.common.util.OpenApiResponse;
import com.jz.dm.common.util.stream.HttpDownload;
import com.jz.dm.gateway.OpenApiService;
import com.jz.dm.models.domian.ApiAuth;
import com.jz.dm.models.domian.ApiInterface;
import com.jz.dm.models.domian.ApiInterfaceCustom;
import com.jz.dm.service.ApiInterfaceService;
import com.jz.dm.service.AuthService;
import com.jz.dm.web.annotation.AccessLimit;
import com.jz.dm.web.annotation.ApiLogAspect;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.List;
import java.util.Map;

/**
 * @author ZC
 * @PACKAGE_NAME: com.jz.dm.service.request
 * @PROJECT_NAME: jz-dm-parent
 * @NAME: ApiQueryService
 * @DATE: 2020-12-23/18:10
 * @DAY_NAME_SHORT: 周三
 * @Description:
 **/
@Service("apiQueryService")
@Slf4j
public class ApiQueryService extends ApiParamVerify implements OpenApiService {

    @Override
    public String getOpenApiMethod() {
        return "query";
    }

    @Override
    public String getOpenApiVersion() {
        return "v1.0.0";
    }

    @Autowired
    private ApiInterfaceService apiInterfaceService;
    @Autowired
    private AuthService authService;
    @Autowired
    private RedisUtils redisUtils;
    @Autowired
    private HttpsUtils httpsUtils;

    /**
     * 数据银行扣款链接
     */
    @Value("${data.bank.balanceUrl}")
    private String balanceUrl;


    /**
     * API请求逻辑处理
     *
     * @param request
     * @param response
     */
    @Override
    @ApiLogAspect(description = "API请求日志")
    @AccessLimit(limit = 10000, sec = 1)
    @SentinelResource(value = "api.gateway", fallback = "fallbackGateway")
    @Transactional(rollbackFor = Exception.class, propagation = Propagation.REQUIRED)
    public void doService(OpenApiRequest request, OpenApiResponse response) {
        boolean bResult = false;
        JSONObject parameter = JSONObject.parseObject(request.getOpenApiParams());
        ApiInterface apiInterface = apiInterfaceService.getApiInfo(request.getApiKey());
        ApiAuth apiAuth = null;
        JSONObject jsonParams = null;
        try {
            String reqParams = parameter.getString("reqParams");
            Map paramMap = null;
            if (StringUtils.isNotBlank(reqParams)) {
                jsonParams = JSONObject.parseObject(reqParams);
                paramMap = JSONObject.parseObject(jsonParams.getString("request_fileds"), Map.class);
                jsonParams.put("is_test",false);
            }
            String authCode = parameter.getString("authCode");
            if (StringUtils.isBlank(authCode)) {
                throw new GatewayException(GatewayResultCode.ILLEGAL_REQUEST);
            }
            verifyApiInterface(apiInterface);
            apiAuth = authService.getAuthInfo(authCode);
            verifyAuth(apiAuth);
            //取出缓存数据
            /* String redisReqParam = redisUtils.get(request.getApiKey());*/
            String redisReqParam = null;
            if (StringUtils.isNotBlank(redisReqParam)) {//redis中存在
                //解析出API制作成功时的参数配置
                JSONObject jsonObject = JSONObject.parseObject(redisReqParam);
                String targetUrl = jsonObject.getString("targetUrl");
                String outputType = jsonObject.getString("outputType");
                String joinType = jsonObject.getString("joinType");
                bResult = rangRequestTarget(outputType, targetUrl,  paramMap,
                        jsonParams, joinType, apiInterface, response);
            } else {//不存在查询数据库
                bResult = rangRequestTarget(apiInterface.getOutputType(),
                        apiInterface.getTargetUrl(),  paramMap, jsonParams, apiInterface.getJoinType(), apiInterface, response);
            }
            //调用成功请求数据银行扣款
            if (AuthModeEnum.POWER_CALL_MODE.name().equals(apiAuth.getAuthMode())) {
                notifierMinusMoney(parameter, bResult);
                //按次调用时处理(处理为已调用)
                authService.updateApiAuthStatus(apiAuth);
            }
        } catch (Exception ex) {
            if (ex instanceof GatewayException) {
                //调用失败回调数据银行解冻
                if (AuthModeEnum.POWER_CALL_MODE.name().equals(apiAuth.getAuthMode())) {
                    notifierMinusMoney(parameter, bResult);
                }
                throw (GatewayException) ex;
            }
            log.error("~~~~~~~~~~~~~~~请求api信息异常~~~~~~~~~~~~~");
            log.error("异常信息:{}", ex.getMessage());
            response.setCode(GatewayResultCode.ILLEGAL_REQUEST.getCode());
            response.setMsg(GatewayResultCode.ILLEGAL_REQUEST.getMsg());
        }
    }


    /**
     * 远程请求获取数据
     *
     * @param outputType
     * @param targetUrl
     * @param param
     * @param joinType
     * @param response
     */
    private boolean rangRequestTarget(String outputType, String targetUrl,
                                      Map<String, String> param, JSONObject jsonParams, String joinType,
                                      ApiInterface apiInterface, OpenApiResponse response) {
        if (StringUtils.isBlank(outputType)) {
            outputType = ApiInfoOutTypeEnum.JSON.name();
        }
        if (StringUtils.isBlank(targetUrl)) {
            throw new GatewayException(GatewayResultCode.REQUEST_PARAM_EMPTY);
        }
        if (StringUtils.isBlank(joinType)) {
            throw new GatewayException(GatewayResultCode.API_TYPE_ERROR);
        }
        if ("10002".equals(joinType) || "10009".equals(joinType)) {//数据表查询
            return dataTableSelect(outputType, targetUrl, param, jsonParams, apiInterface, response);
        } else if ("10004".equals(joinType) || "10008".equals(joinType)) {//三方查询
            return thirdSelect(targetUrl, apiInterface, response);
        } else if ("10005".equals(joinType) || "10007".equals(joinType)) {//数据包查询
            return dataBagDownload(targetUrl,jsonParams,response);
        } else {
            throw new GatewayException(GatewayResultCode.API_TYPE_ERROR);
        }
    }

    /**
     * 三方查询 $$ 数据银行+DMP
     *
     * @param targetUrl
     * @param apiInterface
     * @param response
     * @return
     */
    private boolean thirdSelect(String targetUrl, ApiInterface apiInterface, OpenApiResponse response) {
        if ("POST".equalsIgnoreCase(apiInterface.getReqType())) {
            return callMethodResponse(httpsUtils.submitPost(targetUrl, null), response);
        } else {
            return callMethodResponse(httpsUtils.doGet(targetUrl, null), response);
        }
    }

    /**
     * 数据包下载 $$ 数据银行+DMP
     * @param targetUrl
     * @param param
     * @return
     */
    private boolean dataBagDownload(String targetUrl,JSONObject param,OpenApiResponse response) {
        String fileLocation = param.getString("file_location");
        String datasourceId = param.getString("datasourceId");
        if (null == fileLocation || null == datasourceId) {
            throw new GatewayException(GatewayResultCode.DATA_BIG_ADDR_UNEXIST);
        }
        JSONObject requestParams = new JSONObject();
        requestParams.put("file_location",fileLocation);
        requestParams.put("datasourceId",datasourceId);
        HttpDownload.postDownload(targetUrl,requestParams);
        response.setCode(GatewayResultCode.SUCCESS.getCode());
        response.setMsg(GatewayResultCode.SUCCESS.getMsg());
        return true;
    }

    /**
     * 数据表查询
     *
     * @param outputType
     * @param targetUrl
     * @param param
     * @param apiInterface
     * @param response
     * @return
     */
    protected boolean dataTableSelect(String outputType, String targetUrl, Map<String, String> param,
                                      JSONObject jsonParams, ApiInterface apiInterface, OpenApiResponse response) {
        ApiInterfaceCustom apiCustomInfo = checkParamLegal(param, apiInterface);
        if (ApiInfoOutTypeEnum.FLOW.name().equals(outputType)) {//文件流形式请求
            return flowRequestMethod(targetUrl, param, response, apiCustomInfo);
        } else if (ApiInfoOutTypeEnum.JSON.name().equals(outputType)) { //json格式请求
            return jsonRequestMethod(targetUrl, param, jsonParams, response, apiCustomInfo);
        } else {
            throw new GatewayException(GatewayResultCode.OUTPUT_TYPE_EXCEPTION);
        }
    }

    /**
     * 校验参数合法性
     *
     * @param param
     * @param apiInterface
     * @return
     */
    private ApiInterfaceCustom checkParamLegal(Map<String, String> param, ApiInterface apiInterface) {
        ApiInterfaceCustom apiCustomInfo = apiInterfaceService.getApiCustomInfo(apiInterface.getId());
        boolean tag = false;
        if (null != apiCustomInfo) {
            List<Map> mapList = JSONObject.parseArray(apiCustomInfo.getRequestParam(), Map.class);
            if (CollectionUtils.isNotEmpty(mapList)) {
                for (Map map : mapList) {
                    if ((Boolean) map.get("required")) {
                        String name = (String) map.get("name");
                        String field = param.get(name);
                        if (null == field) {
                            tag = true;
                        }
                    }
                }
                if (tag) {
                    throw new GatewayException(GatewayResultCode.REQUEST_PARAM_EMPTY);
                }
            }
        }
        return apiCustomInfo;
    }

    /**
     * 数据查询---flow流请求方式
     *
     * @param targetUrl
     * @param param
     * @param response
     * @return
     */
    private boolean flowRequestMethod(String targetUrl, Map<String, String> param,
                                      OpenApiResponse response, ApiInterfaceCustom apiCustomInfo) {
        String dataSize = param.get("dataSize");
        Integer valueOf = -1;
        if (StringUtils.isNotBlank(dataSize)) {
            valueOf = Integer.valueOf(dataSize);
        }
        String request$Filed =null;
        net.sf.json.JSONObject reqParams = net.sf.json.JSONObject.fromObject(param);
        try {
            request$Filed = URLEncoder.encode(reqParams.toString(), "UTF-8");
        } catch (UnsupportedEncodingException e) {
            log.error("流下载入参编码异常-------",e.getMessage());
        }
        JSONObject requestParams = new JSONObject();
        requestParams.put("datasourceId",apiCustomInfo.getEsDataSource());
        requestParams.put("query_database",apiCustomInfo.getEsDataBase());
        requestParams.put("query_table",apiCustomInfo.getEsTable());
        requestParams.put("request_fileds",request$Filed);
        requestParams.put("response_fields",assembleResponseParams(apiCustomInfo.getResponseParam()));
        requestParams.put("data_size",valueOf);
        HttpDownload.postDownload(targetUrl,requestParams);
        response.setCode(GatewayResultCode.SUCCESS.getCode());
        response.setMsg(GatewayResultCode.SUCCESS.getMsg());
        return true;
    }
    /**
     * 数据查询--json请求方式
     *
     * @param targetUrl
     * @param param
     * @param response
     * @param apiCustomInfo
     * @return
     */
    private boolean jsonRequestMethod(String targetUrl, Map<String, String> param, JSONObject jsonParams,
                                      OpenApiResponse response, ApiInterfaceCustom apiCustomInfo) {
        JSONObject params = new JSONObject();
        try {
            params.put("datasourceId", apiCustomInfo.getEsDataSource());//数据源id
            params.put("query_database", apiCustomInfo.getEsDataBase());//数据源库名称
            params.put("query_table", apiCustomInfo.getEsTable());//数据源库表
            net.sf.json.JSONObject reqParams = net.sf.json.JSONObject.fromObject(param);
            params.put("request_fileds", reqParams);//请求参数
            params.put("response_fields", assembleResponseParams(apiCustomInfo.getResponseParam()));//响应参数
            params.put("is_test",jsonParams.get("is_test"));//是否是测试
            Integer pageNum = jsonParams.getInteger("page_num");
            Integer pageSize = jsonParams.getInteger("page_size");
            params.put("page_size", apiCustomInfo.getPageSize());
            params.put("page_num", apiCustomInfo.getPageNum());
            if (null != pageNum) {
                params.put("page_num", pageNum);
            }
            if (null != pageSize) {
                params.put("page_size", pageSize);
            }
        } catch (Exception ex) {
            log.error("数据转换异常:{}", ex.getMessage());
            ex.printStackTrace();
        }
        String respResult = httpsUtils.submitPost(targetUrl, params.toString());
        return callMethodResponse(respResult, response);
    }

    /**
     * 组装响应参数
     *
     * @param respParams
     * @return
     */
    private String assembleResponseParams(String respParams) {
        StringBuilder builder = new StringBuilder();
        List<Map> mapList = JSONObject.parseArray(respParams, Map.class);
        if (CollectionUtils.isNotEmpty(mapList)) {
            for (Map map : mapList) {
                String name = (String) map.get("name");
                builder.append(name).append(LoggingConstants.SEP);
            }
        }
        return builder.substring(0, builder.length() - 1);
    }

    /**
     * 调用方法处理结果
     *
     * @param result
     * @param response
     */
    private boolean callMethodResponse(String result, OpenApiResponse response) {
        if (null == result) {
            throw new GatewayException(GatewayResultCode.DISTANCE_REQUEST_EXCEPTION);
        }
        JSONObject resp = JSONObject.parseObject(result);
        if (null != resp && "ESC00000".equals(resp.getString("return_code"))) {
            response.setCode(GatewayResultCode.SUCCESS.getCode());
            response.setMsg(GatewayResultCode.SUCCESS.getMsg());
            response.setAttribute(resp);
            return true;
        } else {
            log.error("~~~~~~~~~~~~~~~~远程请求异常~~~~~~~~~~~~~~~~~");
            throw new GatewayException(GatewayResultCode.DISTANCE_REQUEST_EXCEPTION);
        }
    }

    /**
     * 通知扣款
     *
     * @param parameter
     * @param bResult
     */
    private void notifierMinusMoney(JSONObject parameter, boolean bResult) {
        Integer assetsId = parameter.getInteger("assetsId");
        Integer userId = parameter.getInteger("userId");
        String dataPrice = parameter.getString("dataPrice");
        JSONObject jsonReq = new JSONObject();
        jsonReq.put("assetsId", assetsId);
        jsonReq.put("userId", userId);
        jsonReq.put("dataPrice", dataPrice);
        jsonReq.put("callStatus", bResult);//true 调用成功 扣款 false 调用失败,解冻金额
        String responseResult = httpsUtils.submitPost(balanceUrl + "/mall/financeCustomerAssets/unfreezeMoney", jsonReq.toString());
        JSONObject paramsResult = JSONObject.parseObject(responseResult);
        if (null != paramsResult) {
            if (200 != paramsResult.getInteger("code")) {
                log.info("~~~~~~~~~~~~~~~调用数据银行扣款失败~~~~~~~~~~~~~");
                throw new GatewayException(GatewayResultCode.CALL_AMOUNT_NOT_ENOUGH);
            }
        }
    }

    /**
     * 限流返回方法
     *
     * @param request
     * @param response
     */
    public void fallbackGateway(OpenApiRequest request, OpenApiResponse response) {
        log.info("用户请求过于频繁触发限流接口:ApiKey为:{}" + request.getApiKey());
        throw new GatewayException(GatewayResultCode.RATE_LIMIT_EXCEEDED);
    }
}
